#!/usr/bin/env python

VER=3.0

"""
# disass v3.0 by atlas
#
# Syntax:  disass3 <binary-executable>
#
# This is the complete rewrite of Disass Disassembly tool.
# 
# Architecture layout:
#
# disass-cli     - main .  Bootstraps the thing
# VirtualMemory  - Class which holds the entire virtual memory for this binary.
#                  Disassembly code in here
#                  Subs go here
#                  Sections go here
#                  Pheaders, Directories, etc... go here   (OTHER)
# FileParser     - Base Executable code.  Extended by ELF, PE, and MACH-O handlers
#   disassELF    - Subclass of FileParser which handles the ELF specifics
#                  (section-loading, lookup-table population, etc)
#                  (handling relocs/GOT/stuff into lookups)
#   disassPE     - Subclass of FileParser which handles the PE specifics
# Section        - Represents a Binary Section.  
#  ELFPheader    - Represents a Program Header
#  PEDirectory
# Sub            - Class for holding a sub's info.  
# Opcode         - Generic Opcode object which is extended by implementation objects
#  disassLibDisassemble- Official disass disassembler.
#  disassPydasm  - Impl of Opcode which uses an existing install of pydasm.
#                  disassPydasm is unsupported as it does not scale well and does not pickle.
#                  Once a binary crosses the "MAGIC_DATABASE_SIZE_TRIGGER", disassPydasm will not work at all.
#                  If you use disassPydasm, set MAGIC_DATABASE_SIZE_TRIGGER to an amazingly high number like 4000000000
#                  so it *never* uses the database backend.  In the future, this back-end will be required, as disass 
#                  starts to save state information.  Unless pydasm becomes pickleable before then, this module will cease
#                  to function.
# 
# disassGUI is possible.  Making disass interactive opens so many doors to improved reversing.  Either pygtk or pyqt are on my options list.


Changelog:
    3.0 
     Complete rewrite using modular file-parsers and modular disassembly library.  
     Added database backend for large files.  Only opcode objects are stored in the database, although other objects should follow soon
    2.3 
      More improvement on data lookups and representation
    2.2:
     Improved data interpretation and display (using dataCHECK technology)
     Altered unnamed_subs to use ADDRESS instead of sub #.
    2.1:
     Data Section Handling
    2.0:
     Changed design to OOP
     Enhanced Sub tagging, including naming
     Added Variable tagging
     
    1.1:
     Translated to Python
     Added 
     Bugfixes
     
    1.0:
     Initial version in Perl


# Read in executable
#	check for binary signature
#	initialize DisassELF or DisassPE accordingly
# Split into sections (ELF/PE)
# Look through Sections for non-executable sections (Data)
#	Prepopulate lookup table
# Parse X sections / Program Headers
#	Look for Sub splits
#		stack frame, ret, 

"""
OS_TEMP_DIR_POSIX = '/var/tmp'
OS_TEMP_DIR_WINDERS = 'C:\Windows\Temp'
OS_TEMP_DIR = None

import sys
import os
if os.name == 'nt':
    OS_TEMP_DIR = OS_TEMP_DIR_WINDERS
else:
    OS_TEMP_DIR = OS_TEMP_DIR_POSIX

try:
    ############  psyco optimization.  If you do not use psyco, comment this out ###############
    import psyco
    psyco.full()
    psyco.log("%s%s%s"%(OS_TEMP_DIR, os.sep, "disass-psyco.log"))  # User likely doesn't have write access to the bin dir
    ############
except:
    print >>sys.stderr, ("'psyco' not found, disabling psyco optimization.")
    #x,y,z = sys.exc_info()
    #sys.excepthook(x,y,z)

import re
from disass3 import *
from disass3.fileparsers import *
import disass3.local.disassemble as disassemble
from atlasutils.smartprint import *

import bsddb
from bsddb import dbshelve

##### CAN ONLY EVER IMPORT ONE OF THESE
#from disassPyDasm import *
from disassLibDisassemble import *


MAGIC_DATABASE_SIZE_TRIGGER = 1000000		# opcodes are stored in (filesize < TRIGGER) ? dict : bsddb.dbshelve

# fileparser modules will load here (look to the end of the file or in disass-cli)
global FILE_PARSERS
FILE_PARSERS = []

# These regex on the "current opcode bytes" will trigger a kick to a new subroutine
SUB_PROLOGS = [ '^\x55.{0,10}\x89\xe5', '^\x55.{0,10}\x8b\xec' ]



## printSyntax() does just that. ##
def printSyntax():
    print >>sys.stderr, """
disass v%0.2f Enhanced Disassembler

Syntax:  %s [--txt|--html] [-S "section1,section2,etc.."] <binary-to-disass>
        -S is the list of sections to disassemble in addition to '.text'

"""  % (VER, sys.argv[0])

class disass:
    """ disass is the glue that holds it all together.  
    Not really doing much except loading the various FileParsers, determining which is best, and bootstrapping the VirtualMemory object.
    """
    def __init__(self, binary, name=None):
        try:
            self.name = name
            self.bytes = binary
            try:
                self.bytes = file(binary).read()
                self.name = binary
            except IOError, e:
                pass
            
            self.EXE = None
            self.memory = None
            self.init()
        except:
            errorhandler()

    def init(self):
        """
        init() finishes off where __init__() leaves off.  __init__ only loads self.bytes"""
        for parser in FILE_PARSERS:
            if (parser.Parser.validate(None, self.bytes)):
                self.EXE = parser.Parser(self.bytes, self.name)
        self.memory = VirtualMemory(self.EXE)

        
    def __str__(self):
        ### REALLY rudimentary.  Make this better as we go.
        output = "disass v%0.2f Enhanced Disassembler\n" % VER
        output += str(self.EXE) + "\n"
        output += str(self.memory)
        return output

FPMAXSTRINGLEN = 100
FPLOOKUPLEN = 20
DATAHEXCOEFF = 20


class FileParser:
    """ FileParser is a base class for ELF And PE and MACHO parsing classes to subclass from.  These subclasses should be .py files in 
        the disass/fileformats/
        What do we need from a FileFormat parser for disass?
        * sections
        * codeSections()        (returns list of sections which are Executable)
        * binarySections()      (returns list of sections which meet certain binary:ascii  ratio...  Perhaps non-exe memory is copied somewhere?
        * lookups
        * getBytes()
        * starting address
        * header information
        * program headers?????  How does this help me?  Could this be summed up in header information?
        * """
    def __init__(self, binary, name = None):
        self.binary = binary
        self.LOOKUPS = {}
        self.startaddress = 0
        self.endaddress = 0
        self.endcode = 0
        self.filetype = "STUPID"
        self.parser = None
        self.name = name
        self.globalcount = 0
        self.init()
        
    def lookup(self, address, unknown=None, findoffset=True):
        """ lookup() implements lookup caching then hands off the format-specific lookups.
          address is the address (str) to resolve.  any 0x, $, * or ' ' chars will be stripped
          unknown indicates how you wish for unknown addresses to be returned
          findoffset indicates whether to fuzzy search... ie. some location within a subroutine would be sub+offset
      the FileParser will want to prefill any lookups it can to save time while processing.  That will allow  FPlookup to do more logical lookups, not simply checking for stuff that should be in the cache.  
      
      Right now I'm not sure if it is better to store string keys or numerical.  Let's go with numerical for now...  we can do a case-study when everything else is working. """
        ### addy = chopaddress(address)  ## Not necessary for numerical keys
        ### Check self.LOOKUPS for the address ###
        #print >>sys.stderr,(addy)
        name = self.LOOKUPS.get(address, None)
        if (not name):
            name = self.FPlookup(address, findoffset)
            if (name):
                self.LOOKUPS[address] = name
            else:                   # If STILL not name
                if (address > self.endcode):
                    self.globalcount += 1
                    name = "Global_%d_%x"%(self.globalcount, address)
                    self.submitLookup(address, name)
                    
        return name
            
    def submitLookup(self, key, value):
        """ submitLookup() allows the addition of lookups to the cache."""
        if self.LOOKUPS.get(key):
            self.LOOKUPS[key] = "%s, %s"%(self.LOOKUPS[key], value)
        else:
            self.LOOKUPS[key] = value
    
            
    def printData(self):
        output = "\nLookup Table and Data\n"
        lastitemaddress = self.startaddress
        keys = self.LOOKUPS.keys()
        keys.sort()
        for i in keys:
            preface=" "
            output += "%s%x: %s\n"%(preface, i, self.LOOKUPS[i])
        return output
    
    def setName(self, name):
        self.name = name
        
    def getName(self):
        return self.name
        
    ### The following methods should be overridden by implementations
    def init(self):
        """ init is the placeholder to be overwritten by implementations """
        pass
    def validate(self, bytes):
        """ validate is a placeholder to be overwritten by implementations
        The purpose of validate is to see if, given a set of bytes, this parser knows how to deal with this file type
        For instance, the ELF parser will be looking for '\x7fELF' and the PE parser will likely be looking for "MZ" """
        pass

    def FPlookup(self, address, findoffset=True):
        """ FPlookup is separate from lookup() to make it easier for Parsers to override the lookup but not the caching.
        FPlookup is the lookup for data not preloaded into the CACHE
        If the address is still not found, returns LOOKUP_NOTFOUND """
        # Scan through sections looking for the one which contains this address
        if address < self.startaddress:
            return None
        for sec in self.getDataSections():
            if (address >= sec.startaddress and address < (sec.endaddress)):
                offset = address - sec.startaddress
                b = sec.getBytes()
                try:
                    numeric = int(b[offset:offset+4],16)
                    if self.getSectionByAddr(numeric) != None:
                        return "0x%.8x"%numeric
                except:
                    pass

                lentonull = b.find('\x00', offset+1)
                if lentonull > FPMAXSTRINGLEN or lentonull < 2:
                    lookup = b[offset:offset + FPLOOKUPLEN]
                else:
                    lookup = b[offset:offset + lentonull]

                if checkDATA(lookup, DATAHEXCOEFF):
                    try:
                        x = offset + 1
                        while (b[x] != '\x00'):
                            x +=  1
                    except:
                        pass

                    lookup = "'%s' (%s+0x%x)"%(b[offset:x], sec.name, offset)

                else:
                    lookup = "'%s'... (%s+0x%x)"%(hexText(lookup), sec.name, offset)
                return lookup
        return None

    def getSubName(self, address, section = '.text'):
        """ getSubName() is called for each new subroutine to get its name.  This may be from a lookup table or generated on the fly.
        This is a simple subroutine, and may be preempted my more useful naming scheme, for instance based on interrupts and external functions called, etc... """
        pass
    
    def getSections(self):
        """ getSections() should return a list of disass.Section """
        pass

    def getCodeSections(self):
        """ getCodeSections(), as you may have guessed, returns a list of only EXECUTABLE Sections"""
        sections = []
        for sec in self.getSections():
            if (sec.isCode() and sec.getFsize() > 0):
                sections.append(sec)
        return sections

    def getDataSections(self):
        sections = []
        for sec in self.getSections():
            if (sec.isData() and sec.getFsize() > 0):
                sections.append(sec)
        return sections

    def getHeaderSummary(self):
        """ getHeaders() should return a list of output for display
        This should include the GOT table, RELOCs, PLT, IMPORTs, Directory Contents, etc... """
        pass
    def __str__(self):
        pass
    def getBytes(self):
        pass
    
    
    
############### VIRTUAL MEMORY CONTAINER CLASS #################

VIRTUALMEMORYTYPE_VMROOT = -1
VIRTUALMEMORYTYPE_CODE_SECTION = 0
VIRTUALMEMORYTYPE_DATA_SECTION = 1
VIRTUALMEMORYTYPE_PHEADER = 2

class VirtualMemoryContainer:
    """ VirtualMemoryContainer is an abstraction of Sections and Program Headers, Directories, etc...  
    This will allow us to construct Virtual Memory for both ELF and PE binaries.
    Subclasses should include:
        PHeader
        Section
        Directory
        etc....
        
    It should be NOTED that all binary vars will point to the same str object, and that all objects referencing a binary 
    are to use offsets to get to the data.  In the end, this should prove a significant memory management benefit.  If it is 
    too slow, we'll consider altering course later.  Subs will have their list of opcodes, Lookups will be initially held in a cache
    """
    def __init__(self, parent = None, name=None, offset=0, startaddress=0, endaddress=0, type=None, flags=0, section = None):
        self.next = None
        self.prev = None
        self.binary = None
        self.name = name
        self.startaddress = startaddress
        self.endaddress = endaddress
        self.offset = offset
        self.flags = flags
        self.type = type
        self.setParent(parent)  # This could be another VirtualMemoryContainer or a VirtualMemory object    
        self.children = {}
        self.childkeys = []
        
    def setParent(self, parent):
        self.parent = parent
        if (parent.binary):
            self.binary = parent.binary
    def addChild(self, key, child):
        self.children[key] = child
        #if (not self.childkeys contains key):   # keys should be indexed after all children are added.
            #self.childkeys.append(key)
            #self.childkeys.sort()  (sorting should be done before access, not each addition.
            
    def indexChildren(self):
        self.childkeys = self.children.keys()
        self.childkeys.sort()
        
    def __len__(self):
        return self.endaddress - self.startaddress
    
    def lookup(self, address, looker=None):
        return self.parent.lookup(address, looker)
    
    def getBytes(self):
        pass
    
    def addMemRef(self, dst, src, sub):
        self.parent.addMemRef(dst, src, sub)
        
    def addCaller(self, dst, src, sub):
        self.parent.addCaller(dst, src, sub)
        
    def addJMPer(self, dst, src, sub):
        self.parent.addJMPer(dst, src, sub)
    
    def getSub(self, address):
        return self.parent.getSub(address)
    
    def getSubName(self, address):
        return self.parent.getSubName(address)
        
    def getOwnerSub(self, address):
        return self.parent.getOwnerSub(address)

    
############### SECTION CLASS ################

class Section(VirtualMemoryContainer):
    """ Section is a wrapper for ELF Section or PE Section.  It is used to abstract necessary Section information for use in disass.  Not sure yet if we'll use this for Program Headers... if we even include Program Headers.  """    
    def __init__(self, name = "unset", EXEsection = None, startaddress = 0, endaddress = 0, fsize = 0, msize = 0):
        self.name = name
        self.offset = 0
        self.startaddress = startaddress
        self.endaddress = endaddress
        self.fsize = fsize
        self.msize = msize
        self.EXEsection = EXEsection
        self.children = {}
        self.childkeys = []
        
    def __len__(self):
        return int(self.msize)
        
    def isData(self):
        return (self.type == VIRTUALMEMORYTYPE_DATA_SECTION)
    
    def isCode(self):
        return (self.type == VIRTUALMEMORYTYPE_CODE_SECTION)
    
    def getBytes(self):
        """ getBytes() calls to the EXEsection so that incidental oddities can be handled there.  For instance, zlib-compressed sections, in the binary are useless, and only after decompressing them can they be disassembled.  Yes, this is a reality I've run into. """
        return self.EXEsection.getBytes()
    
    def getOffset(self):
        return self.offset
    
    def getstartaddress(self):
        return self.startaddress
    
    def getFsize(self):
        return self.fsize
    
    def __str__(self):
        output = "\n\n\n\n====== Section Header:       %s ======\n======  Starting Address: \t0x%.8x\tLength:      \t0x%.4x\n======  File Offset:    \t0x%.4x\t\tSize in File:\t0x%.4x\n======  Size in Memory:   \t0x%.4x\n==================\n"%(self.name, self.startaddress, self.endaddress-self.startaddress, self.offset, self.fsize, self.msize)
        return output




############### VIRTUAL MEMORY CLASS #################

class VirtualMemory(VirtualMemoryContainer):
    """ VirtualMemory is a special type of VirtualMemoryContainer, specifically it is the root of Virtual Memory
    and the main interface to outside APIs.
    If "exe" is set then you don't need "binary".  In fact, "binary" will be overwritten with exe.getBytes() """
    def __init__(self, exe=None, name=None, binary=None, offset=0, startaddress=0, endaddress=0, type=None, flags=0):
        self.next = None
        self.prev = None
        self.name = name
        self.startaddress = startaddress
        self.endaddress = (startaddress,0)[startaddress>endaddress]
        self.offset = offset
        self.odd = []
        self.flags = flags
        self.binary = binary
        self.EXE = exe              # This will hold the FileParser object.
        if (exe):
            self.binary = exe.getBytes()       #### What's the likelihood of any value coming from "binary bytes"... most information comes from sections of program headers
            self.name = exe.getName()
            self.startaddress = exe.startaddress
            self.LOOKUPS = exe.LOOKUPS
            self.lookup = exe.lookup

        self.type = VIRTUALMEMORYTYPE_VMROOT
        self.children = {}
        
        self.addChild('OPS', {})
        #self.addChild('SUBS', openDB(self.name, 'main', "subs"))
        self.addChild('SUBS', {})
        self.addChild('SECTIONS', {})
        self.addChild('OTHER', {})
        if (len(self.EXE) > MAGIC_DATABASE_SIZE_TRIGGER):
            self.addChild('OPS', openDB(self.name, 'main', 'ops'))
        else:
            self.addChild('OPS', {})
        self.init()
        
    def init(self):
        """ init() takes care of the additional stuff once the VirtualMemory object has been initially initialized :)  This may go away in the future."""
        for sec in self.EXE.getCodeSections():
            print >>stderr, ("VirtualMemory:init:section:  %s"%sec.name)
            opsname = '%s-op'%sec.name		#How it's stored in self.children
            subsname = '%s-sub'%sec.name
            #if (len(sec) > MAGIC_DATABASE_SIZE_TRIGGER or sec.getFsize() > MAGIC_DATABASE_SIZE_TRIGGER):
            #    self.addOPS(opsname, openDB(self.name, sec.name, 'ops'))
            #else:
            #    self.addOPS(opsname, {})

            contempsubs = self.parse(sec.getBytes(), sec.startaddress, sec)

            tmpsec = self.getSECTIONS().get(sec.startaddress, None)
            if tmpsec == None:
                self.getSECTIONS()[sec.startaddress] = [sec]
            else:
                tmpsec.append(sec)
                self.getSECTIONS()[sec.startaddress] = tmpsec
                print >> sys.stderr,(tmpsec)

            sec.children.update(contempsubs)            # This ties all the subroutines of this section back to the section itself, allowing for manip. to be done just based on section 
            
    def __str__(self):
        output = "Virtual Memory: %s, \t start: %x, \t flags: %d\n\n"%(self.name, self.startaddress, self.flags)
        previoussub = 0
        SECTIONS = self.getSECTIONS()
        SECTIONK = SECTIONS.keys()
        OTHER = self.children['OTHER']
        OTHERK = OTHER.keys()
        SUBS = self.getSUBS()
        SUBK = SUBS.keys()
        SECTIONK.sort()
        OTHERK.sort()
        SUBK.sort()
        
        for k in SUBK:
                ### Check to see whether we are or have crossed a boundary (Program Headers, for example)
                if (len(OTHERK) > 0 and OTHER[0] <= k):
                    output += str(OTHER[OTHERK.pop(0)][0])
                ### Check to see whether we are or have crossed a section boundary
                if (len(SECTIONK) > 0 and SECTIONK[0] <= k):
                    output += str(SECTIONS[SECTIONK.pop(0)][0])
                    
                ### But focus on the code...
                output += str(SUBS[k])
            
        output += self.EXE.printData()
        return output
        
    ### VirtualMemory    
    def parse(self, bytes, startaddress, section = '.text'):
        """
        parse() takes in bytes at an address and populates SubRoutines.  parse() expects to be called for each code section
        
        Once all subroutines and markup stores are parsed, process() method should be called on the object.  process() runs through the existing VirtualMemory object and marks up the SubRoutines, associates calls and jmps and memrefs, and possibly autonames subs... or perhaps autoname() can be called separately.  There is no reason it should be run from process or parse.  
        take in bytes
        run through bytes:
            if new subroutine, close out the existing sub, create and attach a new one.
            get opcode and length
            submit a tuple of (address,the opcode, [links], string descriptor) to the appropriate subroutine (which lives off here)
            
        """
        contempsubs = {}
        offset = 0  # this keeps track of how far into bytes we are.
        bytelen = len(bytes)
        while (offset < bytelen):
            #print >>stderr,("VMA-parser: bytelen: %x\t offset: %x\t VMA: %x"%(len(bytes), offset, startaddress+offset))
            sub = Sub(bytes, offset, self.EXE, self, section, startaddress+offset)
            self.getSUBS()[startaddress+offset] = sub
            tmpsub = contempsubs.get(startaddress+offset,None)
            if tmpsub == None:
                contempsubs[startaddress+offset] = [sub]
            else:
                tmpsub.append(sub)
                contempsubs[startaddress+offset] = tmpsub

            #print >>stderr,("    VMA-parser: sublength: %x"%len(sub))
            newlen = len(sub)
            if newlen < 1: 
                print >>stderr,("ERROR: Sub Length < 1:  %d"%newlen)
            offset += newlen
            if (offset % 10000 < newlen):
                print >>stderr, ("%d / %d"%(offset,bytelen))
            
        ### This could be done outside parse, but since we're here....
        for subarry in contempsubs.values():  # Use of contempsubs means we don't rerun through other subs which might have been created during a different run.  The idea is to allow adding code sections later.
            for sub in subarry:
                sub.process()
        
        return contempsubs
      
    def getOPS(self, namespace = None):
        """
        ignore namespace, it is remnant of a failed experiment and possibly future work will use it.
        """
        if namespace:
            return self.children['OPS'].get('%s-op'%(namespace), None)
        else:
            return self.children['OPS']

    def getOp(self, address, namespace = None):
        if namespace:
            return self.getOPS(namespace).get(address, None)
        else:
            for sec in self.getOPS():
                if address in sec:
                    return sec[address]

    def getSUBS(self, namespace = None):
        """
        ignore namespace, it is remnant of a failed experiment and possibly future work will use it.
        """
        if namespace:
            return self.children['SUBS']['%s-sub'%namespace]
        else:
            return self.children['SUBS']

    def getSECTIONS(self):
        return self.children['SECTIONS']

    def getSub(self, address):
        for sub in self.getSUBS().values():
            if sub.contains(address):
                return sub

    def getSubName(self, address):
        name = self.EXE.getSubName(address)
        return name
        
    def getOwnerSub(self, address):
        subs = self.getSUBS()
        for sub in subs.values():
            if sub.contains(address):
                return (sub, address-sub.startaddress)
        return (None, 0)
    
        
    def addMemRef(self, dst, src, sub, section = '.text'):
        op = self.getOPS().get("%x"%dst, None)
        if not op:
            op = disassOpcode(None,0,dst)
            nsub = self.getOwnerSub(dst)
            if (nsub[0]):
                nsub[0].odd.append((dst, src, sub))
            else:
                self.odd.append((dst, src, sub))
        op.addMemRef(src,sub)
        
    def addCaller(self, dst, src, sub, section = '.text'):
        op = self.getOPS().get("%x"%dst, None)
        if not op:
            op = disassOpcode(None,0,dst)
            nsub = self.getOwnerSub(dst)
            if (nsub[0]):
                nsub[0].odd.append((dst, src, sub))
            else:
                self.odd.append((dst, src, sub))
        op.addCaller(src,sub)
                
    def addJMPer(self, dst, src, sub, section='.text'):
        op = self.getOPS().get("%x"%dst ,None)
        if not op:
            op = disassOpcode(None,0, dst)
            nsub = self.getOwnerSub(dst)
            if (nsub[0]):
                nsub[0].odd.append((dst, src, sub))
            else:
                self.odd.append((dst, src, sub))
        op.addJMPer(src,sub)




############### CLASS SUB - Subroutines ################

OPCHUNKLEN = 12
class Sub(VirtualMemoryContainer):
    FORMAT="AT&T"
    
    def __init__(self, bytes, offset = 0, exeparser = None, parent = None, section = None, startaddress = 0):
        self.parent = parent
        self.startaddress = startaddress            # startaddress indicates the first virtual address of this Sub
        self.endaddress = 0                         # endaddress indicates the virtual address where the next Sub would start
        self.baseaddress = startaddress - offset    # baseaddress indicates the address at offset (for calculations of endaddress later)
        self.exeparser = exeparser  # Elf, PE, or Mach Parser object
        self.section = section
        self.bytes = bytes
        self.ops = self.parent.getOPS()
        self.offset = offset        # offset signifies the offset from the start of "bytes" where this Sub begins...
        self.name = ""
        self.lines = []
        self.odd = []
        self.vrbls = []
        #self.callers = []
        #self.jmpers = []
        #self.refers = []
        
        self.parse()

        #self.bytes = ''
        #print >>sys.stderr,("New Sub:  VMA:0x%.08x\t\t File Offset:%x"%(startaddress,offset))
        
    """  UNNECESSARY WITH ALL OPCODES IN ONE LOCATION.....  Directly Update OPS
    ### CLASS SUB
    def addMemRef(self, dst, src, sub):
        self.refers.append((dst, src, sub))
        for line in self.lines:
            if (line.address == dst):
                line.addMemRef(src, sub)
        
    ### CLASS SUB
    def addCaller(self, dst, src, sub):
        self.callers.append((dst, src, sub))
        for line in self.lines:
            if (line.address == dst):
                line.addCaller(src, sub)
        
    ### CLASS SUB
    def addJMPer(self, dst, src, sub):
        self.jmpers.append((dst, src, sub))
        for line in self.lines:
            if (line.address == dst):
                line.addJMPer(src, sub)
    """
        
    ### CLASS SUB
    def contains(self, address):
        if (address >= self.startaddress and address < self.endaddress):
            return True
        return False
    
    ### CLASS SUB
    def parse(self, bytes = None, offset = None, startaddress = None):
        """ parse() takes in (or handles the object's) bytes.  Using offset and address, new opcodes are created and appended to the opcode storage struct (currently a "list" object) 
        """
        if (bytes):
            self.bytes = bytes
            if (offset):
                self.offset = offset
        if (startaddress):
            self.startaddress = startaddress
        
        byteslen = len(self.bytes)
        offset = self.offset
        bytes = self.bytes
        
        #print >>stderr,('Sub:parse: len(bytes): %x\t\t  offset: %x'%(len(bytes), offset))
        ### Cycle through the bytes in this section or chunk.  When an End of Sub is identified or the end of bytes is reached, return.
        ctinue = True
        while ctinue and offset < byteslen:
            #   Create new opcode object
            #   check for new sub bytes    (should we allow a FileParser to define it's own Sub Pro/Epilog?  or should they just update the global search strings?)
            
            for pattern in SUB_PROLOGS:
                if (bytes[offset] == '\x00' and re.match(pattern, bytes[offset+1:offset+20]) != None):
                        offset += 1
                        ctinue = False
                        #print >>stderr,("End-00 %x"%(self.baseaddress+offset))
                        break
                if (bytes[offset] == '\x55' and len(self.lines) > 0 and (re.match(pattern, bytes[offset:offset+20]) != None)):
                        #print >>stderr,('Sub:parse:End of Sub found')
                        ctinue = False
                        #print >>stderr,("End-== %x"%(self.baseaddress+offset))
                        break
            #   Handle PLT's "single-sub" mentality and break it into appropriate subs
            if self.section and self.section.name == '.plt' and (offset % 16 == 0) and len(self.lines) > 0:
                ctinue = False
                break
                
            #   append to the "lines" list...  should we consider using a dict instead?  It could tie the address together as a key... although binary-search would probably be more efficient if we need to find anything after parsing....  We'll start with a list of objects/tuples but consider a dict later.  I think it would be less efficient.
            # This is a tuple of (address, opcode)
            if (ctinue):
                address = self.baseaddress+offset
                #print >>stderr,("offset: %x \t bytes: %s"%(offset, hexText(bytes)))
                opcode = disassOpcode(bytes, offset, address,32)
                #print >>stderr,("%s:  %s"%(self.name,opcode))
                try:
                    oplen = len(opcode)
                    offset += oplen
                except AttributeError, e:
                    if (opcode.opobj == None):
                        opcode = disassOpcode('\x41', 0, address)
                        if byteslen-offset < 20:
                            dispbytes = bytes[offset:]
                        else:
                            dispbytes = bytes[offset:offset+20]
                        opcode.stringdesc = ['ERROR!!  OPCODE OBJECT FAILED TO CREATE!  ALL SUBSEQUENT DISASSEBLY MAY BE WRONG!  Input: %s'%(hexText(dispbytes))]
                        offset += 1
                        print >>stderr,(opcode.stringdesc[-1])
                addy = "%x"%address
                self.lines.append(addy)
                self.ops[addy] = opcode

                #print >>stderr,("Sub:parse:oplen: %x\t\t offset: %x"%(len(opcode),offset))
            #else:
                ##print >>stderr,("Sub:parse:breakout -> End of Sub")
                #break
            
        #   update length and exit
        self.endaddress = self.baseaddress + offset -1
        #print >>stderr,("Start:  %x\t\t End:  %x"%(self.startaddress,self.endaddress))
        self.bytes = None
        bytes = None
    ### CLASS SUB
    def __len__(self):
        return int(self.endaddress - self.startaddress + 1)
    
    def __str__(self):
        retvar = "\n"
        retvar += "\nSubroutine: %s \t %d lines \t 0x%x bytes   " %( self.name,len(self.lines),self.endaddress-self.startaddress+1)
        
        # Print out Local Variables and Sub Parms
        if (len(self.vrbls) > 0):
            retvar += "\nVariables:"
            vrbls = self.vrbls
            vrbls.sort(reverse=True)
            for x in range(len(vrbls)):
                size = 0
                var1 = vrbls[x]
                if (var1 < 0):
                    # The current variable is a local var
                    if (x > 0):
                        var2 = vrbls[x-1]
                    else:
                        var2 = 0
                    if (var2 < 0):
                        size = var2 - var1
                    else:
                        size = 0 - var1
                    var1 = 0x100000000 + var1
                else:
                    # The current variable is a Parameter
                    if (x > 0):
                        var2 = vrbls[x-1]
                    else:
                        var2 = var1 + 4
                    size = var2 - var1
                
                retvar += "\n    %16x (%5x) _____________________________________________________" % (var1,size)

        # Possibly temporary:  list beginning and end addresses:
        if (self.endaddress-self.startaddress > 16):
            retvar += "\nStarting address: %x\t\tEnding address: %x\n"%(self.startaddress,self.endaddress)
            
        
        """ DISCONTINUED UNTIL FURTHER NOTICE.  THIS FUNCTIONALITY HAS BEEN REPLACED INLINE...
        # Print out "Callers" and "JMPers"
        ### caller is a tuple of (dst, src, (sub,offset))
        ### (sub,offset) is a tuple indicating the sub object and how far into the sub the dst is.
        if (len(self.callers) > 0):  
            retvar += "\nCalled By:\n"
            for caller in self.callers:
                retvar += "\t%s:%x->\t%x\n"%(caller[2].getQuickyRef(), caller[1], caller[0])
        
        if (len(self.jmpers) > 0):
            retvar += "\nJMPed into from:\n"
            for jmper in self.jmpers:
                retvar += "\t%s:%x->\t%x\n"%(jmper[2].getQuickyRef(), jmper[1], jmper[0])
        
        if (len(self.refers) > 0):
            retvar += "\nReferred to by:\n"
            for refer in self.refers:
                retvar += "\t%s:%x->\t%x\n"%(refer[2].getQuickyRef(), refer[1], refer[0])
        """
        
        # Print the Disassembly
        for EACH in self.lines:
            line = self.ops[EACH]
            retvar += "\n%-70s" % (line.printout(self.parent))

        retvar += "\nEnd Subroutine: %s \t %d lines \t 0x%x bytes   \n" % (self.name,len(self.lines), self.endaddress-self.startaddress+1)
        return retvar
        
        
    def getQuickyRef(self):
        return self.name
    
    def processVars(self, op):
        #Call to the opcode instance.  This needs to be independent of the opcode parser stuff
        lclvar = op.getLocalVar()
        if (lclvar and not lclvar in self.vrbls):
            self.vrbls.append(lclvar)
        
    def processOpcode(self, op):
        # Search based on getOp1Type(), getOp2Type(self), getOp3Type(self)
        # Look for OPERANDTYPE_IMMEDIATE and OPERANDTYPE_MEMORY
        # Once found, check for isJMP, isCALL, and isMEMREF
        ### REVAMP:  getOp1Type doesn't work for PyDasm correctly
        
        optype = op.getOp1Type()
        if (optype == OPERANDTYPE_IMMEDIATE or optype == OPERANDTYPE_MEMREF):
            number = op.getOp1Address()
            if number:
                number = number[0]
                if number > 0:
                    osub,osuboff = self.parent.getOwnerSub(number)      #Is the number an address inside a known subroutine?
                    ref = self.parent.lookup(number)            #Is the number referenced in the LOOKUPS table or cache?
                    if (osub):
                        op.addLinksTo(number, osub.startaddress)
                        if (op.isJMP()):
                            if (osub == self):
                                self.addJMPer(number, op.address, self.startaddress)
                            else:
                                self.parent.addJMPer(number, op.address, self.startaddress, self.section.name)
                            return
                        elif (op.isCALL()):
                            if (osub == self):
                                self.addCaller(number, op.address, self.startaddress)
                            else:
                                self.parent.addCaller(number, op.address, self.startaddress, self.section.name)
                            return
                        else:
                            if (osub == self):
                                self.addMemRef(number, op.address, self.startaddress)
                            else:
                                self.parent.addMemRef(number, op.address, self.startaddress, self.section.name)
                    elif (ref):
                        op.addStringRef(ref)

        
        optype = op.getOp2Type()
        if (optype == OPERANDTYPE_IMMEDIATE or optype == OPERANDTYPE_MEMREF):
            number = op.getOp2Address()
            if number:
                number = number[0]
                if number > 0:
                    osub, osuboff = self.parent.getOwnerSub(number)      #Is the number an address inside a known subroutine?
                    ref = self.parent.lookup(number)            #Is the number referenced in the LOOKUPS table or cache?
                    if (osub):
                        op.addLinksTo(number, osub)
                        if (osub == self):
                            self.addMemRef(number, op.address, self.startaddress)
                        else:
                            self.parent.addMemRef(number, op.address, self.startaddress, self.section.name)
                    elif (ref):
                        op.addStringRef(ref)
    
        optype = op.getOp3Type()
        if (optype == OPERANDTYPE_IMMEDIATE or optype == OPERANDTYPE_MEMREF):
            number = op.getOp3Address()
            if number:
                number = number[0]
                if number > 0:
                    osub, osuboff = self.parent.getOwnerSub(number)      #Is the number an address inside a known subroutine?
                    ref = self.parent.lookup(number)            #Is the number referenced in the LOOKUPS table or cache?
                    if (osub):
                        op.addLinksTo(number, osub)
                        if (osub == self):
                            self.addMemRef(number, op.address, self.startaddress)
                        else:
                            self.parent.addMemRef(number, op.address, self.startaddress, self.section.name)
                    elif (ref):
                        op.addStringRef(ref)
    
    
        
        
        
        
        
    ### CLASS SUB
    def setName(self,name = None):
        ### Set the name first
        if (name):
            self.name = name
        else:
            #  Find Name among "peer" objects
            self.name = self.parent.getSubName(self.startaddress)
        
        #print >>sys.stderr, ("NAMING SUB: %s"%self.name)
        
    ### CLASS SUB
    def process(self):
        #return
        """ This tells the object that all lines are in so it can run it's post-processing tasks"""
        self.setName()     #Already done
        sys.stderr.write("BEGIN Processing Sub: %s   ....."%self.name)

        #print >>sys.stderr,("processVars() and processMemRefs()")
        for line in self.lines:
            op = self.ops[line]
            self.processVars(op)
            self.processOpcode(op)

            # Only write back to the database if it's changed.
            #if (op.linksto or op.refsfrom or op.jmpsfrom or op.callsfrom or op.stringdesc):
                #self.ops[line] = op
        
        #
        print >>sys.stderr,("END Processing Sub: %s"%self.name)
        





binary=""
findoffset = True
findsyms = True
finddynsyms = True
printopcodes = True
HELP=False





def errorhandler():
    x,y,z = exc_info()
    excepthook(x,y,z)
    

def openDB(filename, dbname, type='op'):
        """ Setup datastore for opcodes.  This may be time-consuming for mid-size binaries, but an absolute must for the big-dogs """
        try:
            return dbshelve.open("%s.%s-%s.dbm"%(filename, dbname, type))
        except bsddb.db.DBAccessError,e:
            import os
            tmpdirname = '%s%s%s'%(OS_TEMP_DIR, os.sep, filename)
            try:
                os.stat(tmpdirname)
            except OSError,oe:
                os.makedirs(tmpdirname)
            return dbshelve.open("%s%s%s-%s.dbm"%(tmpdirname,os.sep,dbname,type))


#if (len(FILE_PARSERS) == 0):

me = None
if (__name__ == "__main__"):
    
    ###### This code loads the fileparsers into the FILE_PARSERS list
    #print os.path.abspath(os.curdir)
    parserdir = "%s%sfileparsers"%(os.sep.join(sys.modules['disass3'].__file__.split(os.sep)[:-1]), os.sep)
    filename = None
    try:
        for filename in  os.listdir(parserdir):
            if (len(filename) > 3 and filename[0] != "_" and filename[-3:] == ".py"):
                pass
                FILE_PARSERS.append(__import__("disass3.fileparsers.%s"%filename[:-3], None, None, "disass3.fileparsers"))
    except:
        print >>sys.stderr,("Error loading parser %s"%filename)
        pass

    ### Testing and Development are helped by the default load...
    if (len(sys.argv) == 1):
        me = disass('/bin/bash')
        print me
    else:
        me = disass(argv[1])
        print me

