"""
Vtrace Debugger Framework

Vtrace is a *mostly* native python debugging framework which
can be used to quickly write programatic debuggers and research
tools.

I'm not known for writting great docs...  but the code should
be pretty straight forward...

This has been in use for over 2 years privately, but is nowhere
*near* free of bugs...  idiosyncracies abound.

==== Werd =====================================================

Blah blah blah... many more docs to come.

Brought to you by kenshoto.  e-mail invisigoth.

Greetz:
    h1kari - eeeeeooorrrmmm  CHKCHKCHKCHKCHKCHKCHK
    Ghetto - wizoo... to the tizoot.
    atlas - *whew* finally...  no more teasing...
    beatle/dnm - come out and play yo!
    The Kenshoto Gophers.
    Blackhats Everywhere.

"""
# Copyright (C) 2007 Invisigoth - See LICENSE file for details
import os
import sys
import code
import types
import struct
import getopt
import signal
import inspect
import traceback

from threading import *

import cobra

remote = None       # If set, we're a vtrace client (set to serverhost)
cobra_daemon = None
port = 0x5656
symbolcache = {} # A list of vtrace symbols key'd by filename
verbose = False


# Order must match format junk
# NOTIFY_ALL is kinda special, if you registerNotifier
# with it, you get ALL notifications.
NOTIFY_ALL = 0          # Get all notifications
NOTIFY_SIGNAL = 1       # Callback on signal/exception
NOTIFY_BREAK = 2        # Callback on breakpoint / sigtrap
NOTIFY_STEP = 3         # Callback on singlestep complete
NOTIFY_SYSCALL = 4      # Callback on syscall (linux only for now)
NOTIFY_CONTINUE = 5     # Callback on continue (not done for step)
NOTIFY_EXIT = 6         # Callback on process exit
NOTIFY_ATTACH = 7       # Callback on successful attach
NOTIFY_DETACH = 8       # Callback on impending process detach       
# The following notifiers are *only* available on some platforms
# (and may be kinda faked out ala library load events on posix)
NOTIFY_LOAD_LIBRARY = 9
NOTIFY_UNLOAD_LIBRARY = 10 
NOTIFY_CREATE_THREAD = 11
NOTIFY_EXIT_THREAD = 12
NOTIFY_MAX = 20

# File Descriptor / Handle Types
FD_UNKNOWN = 0
FD_FILE = 1
FD_SOCKET = 2
FD_PIPE = 3

# Memory Map Permission Flags
MM_READ = 0x4
MM_WRITE = 0x2
MM_EXEC = 0x1
MM_SHARED = 0x08

# Vtrace Symbol Types
SYM_MISC = -1
SYM_GLOBAL = 0 # Global (mostly vars)
SYM_LOCAL = 1 # Locals
SYM_FUNCTION = 2 # Functions
SYM_SECTION = 3 # Binary section
SYM_META = 4 # Info that we enumerate

# Vtrace Symbol Offsets
VSYM_NAME = 0
VSYM_ADDR = 1
VSYM_SIZE = 2
VSYM_TYPE = 3
VSYM_FILE = 4

from vtrace.rmi import *
from vtrace.notifiers import *
from vtrace.breakpoints import *
from vtrace.util import *

class Trace(Notifier):
    """
    The main tracer object.  A trace instance is dynamically generated using
    this and *many* potential mixin classes.  However, API users should *not*
    worry about the methods that come from the mixins...  Everything that is
    *meant* to be used from the API is contained and documented here.
    """
    def __init__(self):
        # For the crazy thread-call-proxy-thing
        # (must come first for __getattribute__
        self.requires_thread = {}
        self.proxymeth = None # FIXME hack for now...
        self.fireTracerThread()
        # The universal place for all modes
        # that might be platform dependant...
        self.modes = {}
        self.modedocs = {}
        self.notifiers = {}

        self.initMode("RunForever", False, "Run until RunForever = False")
        self.initMode("NonBlocking", False, "A call to wait() fires a thread to wait *for* you")
        self.initMode("ThreadProxy", True, "Proxy necissary requests through a single thread (can deadlock...)")
        self.initMode("FastBreak", False, "Do *NOT* add/remove breakpoints per-run, but leave them there once active")
        self.initMode("SingleStep", False, "All calls to run() actually just step.  This allows RunForever + SingleStep to step forever ;)")

        self.fmt = self.getRegisterFormat()
        self.regnames = self.getRegisterNames()
        self.fb_bp_done = False # A little hack for FastBreak mode


        # Set if we're a server and this trace is proxied
        self.proxy = None

        Notifier.__init__(self)

        # We'll just use our own notify interface to catch some stuff
        # (which also more-or-less guarentees we'll be first notified for these)
        self.registerNotifier(NOTIFY_ALL, self)

        # Add event numbers to here for auto-continue
        self.auto_continue = [NOTIFY_LOAD_LIBRARY, NOTIFY_CREATE_THREAD, NOTIFY_UNLOAD_LIBRARY, NOTIFY_EXIT_THREAD]

        self._initLocals()

    def execute(self, cmdline):
        """
        Start a new process and debug it
        """
        if self.isAttached():
            raise Exception("ERROR - Tracer must first be detached before you can execute()")

        pid = self.platformExec(cmdline)
        self.justAttached(pid)
        event = self.platformWait()
        self.platformProcessEvent(event)

    def addIgnoreSignal(self, code, address=0):
        """
        By adding an IgnoreSignal you tell the tracer object to
        supress the notification of a particular type of signal.
        In POSIX, these are regular signals, in Win32, these
        are exception codes.  This is mostly useful in RunForever
        mode because you still need the process to begin running again.
        (these may be viewed/modified by the metadata key "IgnoredSignals")
        FIXME: make address do something.
        """
        self.getMeta("IgnoredSignals").append(code)

    def delIgnoreSignal(self, code, address=0):
        """
        See addIgnoreSignal for a description of signal ignoring.
        This removes an ignored signal and re-enables it's delivery.
        """
        self.getMeta("IgnoredSignals").remove(code)

    def attach(self, pid):
        """
        Attach to a new process ID.
        """
        if self.isAttached():
            self.detach()
    
        self.platformAttach(pid)
        self.justAttached(pid)
        event = self.platformWait()
        self.platformProcessEvent(event)

    def stepi(self):
        """
        Single step the target process ONE instruction (and do
        NOT activate breakpoints for the one step). Also, we 
        don't deliver pending signals for the single step...
        NOTE: there is no CONTINUE event for a single step
        (for efficiency)
        """
        self.requireAttached()
        self.platformStepi()
        event = self.platformWait()
        self.platformProcessEvent(event)

    def run(self):
        """
        Allow the traced target to continue execution.
        (Depending on the mode "Blocking" this will either
        block until an event, or return immediately)
        """
        if self.getMode("SingleStep", False):
            self.steploop()
        else:
            self._doRun()
            self.wait()

    def kill(self):
        """
        Kill the target process for this trace (will result in process
        exit and fire appropriate notifiers)
        """
        self.requireAttached()
        self.requireNotExited()
        # kill may require that we continue
        # the process before it gets processed,
        # so we'll try to run the process until it
        # exits due to the kill
        self.setMode("RunForever", True) # Forever actually means util exit
        if self.isRunning():
            self.platformKill()
        else:
            self.platformKill()
            self.run()

    def detach(self):
        """
        Detach from the currently attached process.
        """
        self.requireAttached()
        self.fireNotifiers(NOTIFY_DETACH)
        self.platformDetach()
        self.attached = False
        self.pid = 0
        self.mapcache = None

    def getPid(self):
        """
        Return the pid for this Trace
        """
        return self.pid

    def getNormalizedLibNames(self):
        """
        Symbols are stored internally based off of
        "normalized" library names.  This method returns
        the list of normalized names for the loaded libraries.

        (probably only useful for writting symbol browsers...)
        """
        return self.resbynorm.keys()

    def getSymsForFile(self, libname):
        """
        Return the entire symbol list for the specified
        filename.
        """
        res = self.getResolverForFile(libname)
        if not res:
            return []
        return res.symList()

    def getSymByName(self, name, libname):
        """
        Return a VSymbol() object for the given name
        from the given binary file.
        
        See the docs on VSymbols() for details (in vtrace.symbase)
        """
        res = self.getResolverForFile(libname)
        if not res:
            return None
        return res.symByName(name)

    def getSymByAddr(self, addr):
        """
        Return the *closest* VSymbol prior to this but ONLY
        from within the symbol resolver for the filename
        which is mapped at that address.

        See the docs on VSymbols() for details (in vtrace.symbase)
        """
        map = self.getMap(addr)
        if map == None:
            return None
        fname = map[3]
        if not fname:
            return None
        res = self.getResolverForFile(os.path.basename(fname))
        if not res:
            return None
        return res.symByAddr(addr)


    def getRegisterByName(self, name):
        """
        Return an int value for the specified
        register by name.
        """
        self.requireAttached()
        regs = self.getRegisters()
        val = regs.get(name, None)
        if val == None:
            raise Exception("ERROR - Unknown register name %s" % name)
        return val

    def setRegisterByName(self, name, value):
        """
        Set a target register by name to the int()
        (or possibly long) value  specified.
        """
        newregs = {}
        newregs[name] = value
        self.setRegisters(newregs)

    def getRegisters(self):
        """
        Return a dict entry of the "user" registers for this process.
        """
        self.requireAttached()
        regs = {}
        reg = self.platformGetRegs()
        siz = struct.calcsize(self.fmt)
        reglist = struct.unpack(self.fmt,reg[:siz])
        for i in range(len(reglist)):
            regs[self.regnames[i]] = reglist[i]

        return regs

    def setRegisters(self, newregs):
        """
        Set any registers specified in the dict newregs to their int values.
        (only the specified registers are effected)
        """
        self.requireAttached()
        curregs = self.getRegisters()
        for key,val in newregs.iteritems():
            curregs[key] = val

        self.saveRegisters(curregs)

    def searchMemory(self, needle):
        """
        A quick cheater way to searchMemoryRange() for each
        of the current memory maps.
        """
        results = []
        for map in self.getMaps():
            # On most platforms, memory you can't
            # read is in kernel space, and you can't
            # read it with a debugger either ;)
            if not map[2] | MM_READ:
                continue
            try:
                results.extend(self.searchMemoryRange(needle, map[0],map[1]))
            except:
                pass # Some platforms dont let debuggers read non-readable mem
        return results

    def searchMemoryRange(self, needle, address, size):
        """
        Search the specified memory range (address -> size)
        for the string needle.   Return a list of addresses
        where the match occurs.
        """
        results = []
        memory = self.readMemory(address, size)
        offset = 0
        while offset < size:
            loc = memory.find(needle, offset)
            if loc == -1: # No more to be found ;)
                break
            results.append(address+loc)
            offset = loc+1 # Skip one past our matcher

        return results

    def readMemory(self, address, size):
        """
        Read memory from address.  Areas that are NOT valid memory will be read
        back as \x00s (this probably goes in a mixin soon)
        """
        self.requireAttached()
        map = self.getMap(address)
        if not map:
            raise Exception("ERROR - Reading from invalid memory area")

        return self.platformReadMemory(long(address), long(size))

    def readMemoryFormat(self, address, fmt):
        """
        Read some target memory and do a python struct unpack
        (returning the resulting tuple)
        """
        size = struct.calcsize(fmt)
        mem = self.readMemory(address, size)
        return struct.unpack(fmt, mem)

    def writeMemory(self, address, bytes):
        """
        Write the given bytes to the address in the current trace.
        """
        self.requireAttached()
        self.platformWriteMemory(long(address), bytes)

    def writeMemoryFormat(self, address, fmt, *args):
        buf = struct.pack(fmt, *args)
        self.writeMemory(address, buf)

    def setMeta(self, name, value):
        """
        Set some metadata.  Metadata is a clean way for
        arbitrary trace consumers (and notifiers) to present
        and track additional information in trace objects.

        Any modules which use this *should* initialize them
        on attach (so when they get re-used they're clean)

        Some examples of metadata used:
        ShouldBreak - We're expecting a non-signal related break
        ExitCode - The int() exit code  (if exited)
        PendingSignal - Used on posix systems
        PendingException - Mostly for win32 for now

        """
        self.metadata[name] = value

    def getMeta(self, name, default=None):
        """
        Get some metadata.  Metadata is a clean way for
        arbitrary trace consumers (and notifiers) to present
        and track additional information in trace objects.

        If you specify a default and the key doesn't exist, not
        not only will the default be returned, but the key will
        be set to the default specified.
        """
        if default:
            if not self.metadata.has_key(name):
                self.metadata[name] = default
        return self.metadata.get(name, None)

    def hasMeta(self, name):
        """
        Check to see if a metadata key exists... Mostly un-necissary
        as getMeta() with a default will set the key to the default
        if non-existant.
        """
        return self.metadata.has_key(name)

    def getMode(self, name, default=False):
        """
        Get the value for a mode setting allowing
        for a clean default...
        """
        return self.modes.get(name, default)

    def setMode(self, name, value):
        """
        Set a mode setting...  This is ONLY valid
        if that mode has been iniitialized with
        initMode(name, value).  Otherwise, it's an
        unsupported mode for this platform ;)  cute huh?
        This way, platform sections can cleanly setmodes
        and such.
        """
        if not self.modes.has_key(name):
            raise Exception("Mode %s not supported on this platform" % name)
        self.modes[name] = bool(value)

    def injectso(self, filename):
        """
        Inject a shared object into the target of the trace.  So, on windows
        this is easy with InjectDll and on *nix... it's.. fugly...
        """
        self.platformInjectSo(filename)

    def ps(self):
        """
        Return a list of proccesses which are currently running on the
        system.
        (pid, name)
        """
        return self.platformPs()

    def addWatchpoint(self, watchpoint):
        """
        Add a "break on access" watch-point provided that the hardware
        has support and sufficient resources are aviailable.  The "watchpoint"
        argument must extend the vtrace.watchpoints.Watchpoint object and will
        have it's "notify" method called when hit...  This method returns
        the ID assigned to the watch point.
        """
        # Watchpoints are handled by the breakpoint subsystem...
        watchpoint.id = self.nextBpId()
        addr = watchpoint.resolveAddress(self)
        if addr == None:
            self.deferredwatch.append(watchpoint)
        else:
            self.archAddWatchpoint(addr)
            self.watchpoints[addr] = watchpoint

    def removeWatchpoint(self, id):
        """
        Remove a previously added watchpoint by id.
        """
        for wp in self.deferredwatch:
            if wp.getId() == id:
                self.deferredwatch.remove(wp)
                return

        for addr,wp in self.watchpoints.items():
            if wp.getId() == id:
                self.watchpoints.pop(addr)
                self.archRemWatchpoint(addr)
                return

    def addBreakpoint(self, breakpoint):
        """
        Add a breakpoint to the trace.  The "breakpoint" argument
        is a vtrace Breakpoint object or something that extends it.

        To add a basic breakpoint use trace.addBreakpoint(vtrace.Breakpoint(address))
        NOTE: expression breakpoints do *not* get evaluated in fastbreak mode

        This will return the internal ID given to the new breakpoint
        """
        breakpoint.id = self.nextBpId()
        addr = breakpoint.resolveAddress(self)
        if addr == None:
            self.deferred.append(breakpoint)
        else:
            if self.breakpoints.has_key(addr):
                raise Exception("ERROR: Duplicate break for address 0x%.8x" % addr)
            self.breakpoints[addr] = breakpoint

        return breakpoint.id
            
    def removeBreakpoint(self, id):
        """
        Remove the breakpoint with the specified ID
        """
        self.requireAttached()
        for bp in self.deferred:
            if bp.getId() == id:
                bp.deactivate(self)
                self.deferred.remove(bp)
                return

        for addr,bp in self.breakpoints.items():
            if bp.getId() == id:
                bp.deactivate(self)
                self.breakpoints.pop(addr, None)

    def getBreakpoint(self,id):
        """
        Return a reference to the breakpoint with the requested ID
        """
        for bp in self.deferred:
            if bp.getId() == id:
                return bp
        for bp in self.breakpoints.values():
            if bp.getId() == id:
                return bp
        return None

    def getBreakpoints(self):
        """
        Return a list of the current breakpoints.
        """
        return self.breakpoints.values() + self.deferred

    def call(self, address, args, convention=None):
        """
        Setup the "stack" and call the target address with the following
        arguments.  If the argument is a string or a buffer, copy that into
        memory and hand in the argument.

        The current state of ALL registers are returned as a dictionary at the
        end of the call...

        Additionally, a "convention" string may be specified that the underlying
        platform may be able to interpret...
        """
        return self.platformCall(address, args, convention)

    def registerNotifier(self, event, notifier):
        """
        Register a notifier who will be called for various
        events.  See NOTIFY_* constants for handler hooks. The 
        notifier method is called with the Trace as it's only arg.
        """
        nlist = self.notifiers.get(event,None)
        if nlist:
            nlist.append(notifier)
        else:
            nlist = []
            nlist.append(notifier)
            self.notifiers[event] = nlist

    def deregisterNotifier(self, event, notifier):
        #FIXME this is broken cause of the proxies...
        nlist = self.notifiers.get(event, [])
        nlist.remove(notifier)

    def getNotifiers(self, event):
        return self.notifiers.get(event, [])

    def requireNotExited(self):
        if self.exited:
            raise Exception("ERROR - Request invalid for trace which exited")

    def requireNotRunning(self):
        """
        Just a quick method to throw an error if the
        tracer is already running...
        """
        if self.isRunning():
            raise Exception("ERROR - Request invalid for running trace")

    def requireAttached(self):
        """
        A utility method for other methods to use in order
        to require being attached
        """
        if not self.attached:
            raise Exception("ERROR - Must be attached to a process")

    def getFds(self):
        """
        Get a list of (fd,type,bestname) pairs.  This is MOSTLY useful
        for HUMON consumtion...  or giving HUMONs consumption...
        """
        self.requireAttached()
        if not self.fds:
            self.fds = self.platformGetFds()
        return self.fds

    def getMaps(self):
        """
        Return a list of the currently mapped memory for the target
        process.  This is acomplished by calling the platform's
        platformGetMaps() mixin method.  This will also cache the
        results until CONTINUE.  The format is (addr,len,perms,file).
        """
        if not self.mapcache:
            self.mapcache = self.platformGetMaps()
        return self.mapcache

    def getMap(self, address):
        """
        Get the mmap for the target address from the current Trace.
        This will return None if the address is not in any map
        (this can also be used as a cheater isMemoryValid() type routine)
        """
        address = long(address)
        maps = self.getMaps()
        for map in maps:
            base = map[0]
            mlen = map[1]
            if address >= base and address < (base+mlen):
                return map

        return None

    def getMapBase(self, filename):
        """
        Return the base map address for the
        memory map which is backed by the given
        file name.
        """
        maps = self.getMaps()
        for map in maps:
            if map[3] == filename:
                return map[0]

        raise Exception("ERROR - Unknown map name %s\n" % filename)

    def isAttached(self):
        """
        Return boolean true/false for weather or not this trace is
        currently attached to a process.
        """
        return self.attached

    def isRunning(self):
        """
        Return true or false if this trace's target process is "running".
        """
        return self.running

    def enableAutoContinue(self, event):
        """
        Put the tracer object in to AutoContinue mode
        for the specified event.  To make all events
        continue running see RunForever mode in setMode().
        """
        if event not in self.auto_continue:
            self.auto_continue.append(self)

    def disableAutoContinue(self, event):
        """
        Disable Auto Continue for the specified
        event.
        """
        if event in self.auto_continue:
            self.auto_continue.remove(event)

    def parseExpression(self, expression):
        """
        This is a convenience wrapper for the VtraceExpression
        object.  It returns the current evaluation of the expression
        right now.  This can raise if invalid syntax or unknown symbols
        are used.

        See the docs for the VtraceExpression object for details on the
        expression syntax...
        """
        e = VtraceExpression(expression)
        return e.evaluate(self)

    def sendBreak(self):
        """
        Send an asynchronous break signal to the target process.
        This is only valid if the target is actually running...
        """
        self.requireAttached()
        if not self.isRunning():
            raise Exception("Why sending a break when not running?!?!")
        self.setMode("RunForever", False)
        self.setMode("FastBreak", False)
        self.setMeta("ShouldBreak", True)
        self.platformSendBreak()

    def getProgramCounter(self):
        """
        An architecture independant way to get the value for the
        program counter on this platform (ie "eip" for all you 
        intel only neordz)  Breakpoint subsystem uses this to be
        independant on my dreamcast...
        """
        self.requireAttached()
        name = self.archGetPcName()
        return self.getRegisterByName(name)

    def setProgramCounter(self, value):
        """
        Set the current instruction pointer.  See getProgramCounter()
        for why this exists.
        """
        self.requireAttached()
        name = self.archGetPcName()
        return self.setRegisterByName(name,value)

    def getStackCounter(self):
        """
        An architecture independant way to get the current stack pointer
        value
        """
        self.requireAttached()
        name = self.archGetSpName()
        return self.getRegisterByName(name)

    def setStackCounter(self, value):
        """
        An architecture independant way to set the current stack pointer
        value
        """
        self.requireAttached()
        name = self.archGetSpName()
        return self.setRegisterByName(name, value)

    def getStackTrace(self):
        """
        Returns a list of (instruction pointer, stack frame) tuples.
        If stack tracing results in an error, the error entry will
        be (-1,-1).  Otherwise most platforms end up with 0,0 as
        the top stack frame
        """
        #FIXME this should be a platform call!
        raise Exception("ERROR - Platform must implement this")

    def getThreads(self):
        """
        Get a dictionary of <threadid>:<tinfo> pairs where
        tinfo is platform dependant, but is tyically either
        the top of the stack for that thread, or the TEB on
        win32
        """
        if not self.threadcache:
            self.threadcache = self.platformGetThreads()
        return self.threadcache

    def selectThread(self, threadid):
        """
        Set the "current thread" context to the given thread id.
        (For example stack traces and register values will depend
        on the current thread context).  By default the thread
        responsible for an "interesting event" is selected.
        """
        if threadid not in self.getThreads():
            raise Exception("ERROR: Invalid threadid chosen: %d" % threadid)
        self.platformSelectThread(threadid)
        self.setMeta("ThreadId", threadid)
        
class TraceGroup(Notifier):
    """
    Encapsulate several traces, run them, and continue to 
    handle their event notifications.
    """
    def __init__(self):
        Notifier.__init__(self)
        self.traces = {}
        self.go = True # A little ghetto switch for those who read the source
        self.notifiers = {}
        self.context = None
        self.modes = {} # See docs for trace modes
        self.metadata = {} # Like traces, but persistant
        self.daemon = None # Each tracegroup gets a daemon # FIXME really?
        self.notifyall = False  # Is this tracegroup just an everythin subscriber?
        self.events = []        # Which events are we subscribing to on tracers?

        for i in range(NOTIFY_MAX):
            self.notifiers[i] = []

        self.mainnotifier = self
        if remote:
            self.daemon = cobra.CobraDaemon(port=0)
            self.daemon.shareObject(self, "TraceGroup")
            self.daemon.fireThread()

        self.setMode("NonBlocking", True)

    def getContext(self):
        return self.context

    def detachAll(self):
        """
        Detach from ALL the currently targetd processes
        """
        for trace in self.traces.values():
            try:
                if trace.isRunning():
                    trace.sendBreak()
                trace.detach()
            except:
                pass

    def subscribeEvent(self, event):
        """
        We only want particular events from the traces
        (this *usually* corresponds to registered notifiers,
        but people who extend from tracegroup and implement
        notify overrides may want to specify specifics)
        """
        if event not in self.events:
            self.events.append(event)
            for trace in self.traces.values():
                notif = self.getGroupNotifier(trace)
                trace.registerNotifier(event, notif)

    def unsubscribeEvent(self, event):
        self.events.remove(event)
        #FIXME for trace in traces, unsubscribe

    def setMeta(self, name, value):
        """
        A trace group's setMeta function will set "persistant" metadata
        which will be added again to any trace on attach.  Additionally,
        setting metadata on a tracegroup will cause all current traces
        to get the update as well....
        """
        self.metadata[name] = value
        for trace in self.traces.values():
            trace.setMeta(name, value)

    def getMeta(name, default=None):
        return self.metadata.get(name, default)

    def run(self):
        """
        Our run method  is a little different than a traditional
        trace. It will *never* block.
        """
        if len(self.traces.keys()) == 0:
            raise Exception("ERROR - can't run() with no traces!")

        for trace in self.traces.values():

            if trace.exited:
                self.traces.pop(trace.pid)
                trace.detach()
                continue

            if not trace.isRunning():
                trace.run()

    def execTrace(self, cmdline):
        trace = getTrace()
        self.initTrace(trace)
        trace.execute(cmdline)
        self.context = trace
        self.traces[trace.getPid()] = trace
        return trace

    def addTrace(self, proc):
        """
        Add a new tracer to this group the "proc" argument
        may be either an long() for a pid (which we will attach
        to) or an already attached (and broken) tracer object.
        """

        if (type(proc) == types.IntType or
            type(proc) == types.LongType):
            trace = getTrace()
            self.initTrace(trace)
            self.traces[proc] = trace
            try:
                trace.attach(proc)
            except:
                self.delTrace(proc)
                raise

        else: # Hopefully a tracer object... if not.. you're dumb.
            trace = proc
            self.initTrace(trace)
            self.traces[trace.getPid()] = trace

        return trace

    def getGroupNotifier(self, trace):
        """
        Create a potentially custom notifier
        for the given trace (only custom for remote)
        """
        if remote:
            csock = cobra.getCobraSocket(trace)
            host,port = csock.getSockName()
            return cobra.CobraProxy("cobra://%s:%d/TraceGroup" % (host, self.daemon.port), retrymax=5)
        return self

    def initTrace(self, trace):
        """
         - INTERNAL -
        Setup a tracer object to be ready for being in this
        trace group (setup modes and notifiers).  Only addTrace()
        probably needs to be aware of this.
        """
        notif = self.getGroupNotifier(trace)
        for event in self.events:
            trace.registerNotifier(event, notif)

        for name,val in self.modes.items():
            trace.setMode(name, val)

        for name,val in self.metadata.items():
            trace.setMeta(name, val)

        self.context = trace

    def delTrace(self, pid):
        """
        Remove a trace from the current TraceGroup
        """
        self.traces.pop(pid, None)
        #FIXME de-register notifiers?

    def getTraces(self):
        """
        Return a list of the current traces
        """
        return self.traces.values()

    def registerNotifier(self, event, notifier):
        """
        Register a notifier which will be added to
        all traces now and in the future.
        """
        if event not in self.events:
            self.subscribeEvent(event)

        nlist = self.notifiers.get(event)
        nlist.append(notifier)

    def deregisterNotifier(self, event, notifier):
        nlist = self.notifiers.get(event, [])
        nlist.remove(notifier)
        #FIXME eventually remove them remote too

    def setMode(self, name, value):
        for trace in self.getTraces():
            trace.setMode(name, value)

        self.modes[name] = value

    def notify(self, event, trace):
        self.context = trace
        # Call our sub-notifiers
        self.fireNotifiers(event, trace)

        # Remove this trace, and free it
        # on the server if present
        if event == NOTIFY_EXIT:
            self.traces.pop(trace.getPid())
            #trace.release() # Let the thread go...
            #if remote: Don't do this quite yet... race conditions...
                ## Tell the remote server we're done with this one
                #factory = getTracerFactory()
                #factory.releaseTrace(trace)

    def fireNotifiers(self, event, trace):
        nlist = self.notifiers.get(NOTIFY_ALL, [])
        for notifier in nlist:
            try:
                notifier.handleEvent(event, trace)
            except:
                print "ERROR - Exception in notifier:",traceback.format_exc()

        nlist = self.notifiers.get(event, [])
        for notifier in nlist:
            try:
                notifier.handleEvent(event, trace)
            except:
                print "ERROR - Exception in notifier:",traceback.format_exc()

class VtraceExpression:
    """
    A vtrace expression is essentially python code with a bunch of metadata
    mapped into the code evaluation namespace.  Currently, every method listed
    in this class except "evaluate" also gets mapped into the space.  This means
    that the expressions may call methods from this object as though they were
    functions ie. the expression:

    mapbase(<address>)

    will call the mapbase method from this object...

    Additionally the following items are mapped directly into the namespace for
    evaluation:

    * Registers
        Mapped in by name

    * Symbol Resolvers 
        They are accessable by the "normalized libname" so, kernel32.dll is "kernel32"
        and /usr/lib/libpthread-2.0.40.so is "libpthread".  These resolvers may be
        dereferenced to get symbols by name.  For example the expression "kernel32.CloseHandle"
        will return the address of the CloseHandle symbol from kernel32

    """
    def __init__(self, expression):
        self.trace = None
        self.expression = expression

    def __str__(self):
        return self.expression

    def evaluate(self, trace, noraise=False):
        """
        Evaluate the expression with metadata/symbols/registers/etc
        from the given tracer.  This *will* raise exceptions on 
        parse errors or invalid symbols etc unless you specify
        noraise at which point it will return None.
        """
        methmap = {
            "frame":self.frame,
            "section":self.section,
            "sectionlen":self.sectionlen,
            "teb":self.teb,
            "poi":self.poi,
            "maplen":self.maplen,
            "mapbase":self.mapbase,
            }

        try:
            locs = {}
            self.trace = trace
            locs.update(trace.getRegisters())
            locs.update(methmap)
            locs.update(trace.resbynorm)
            locs["trace"] = trace
            x = long(eval(self.expression,{},locs))
            self.trace = None
            return x
        except Exception, e:
            self.trace = None
            if noraise:
                if vtrace.verbose: print e
                return None
            raise

    def frame(self, index):
        """
        Return the address of the saved base pointer for
        the specified frame.

        Usage: frame(<index>)
        """
        stack = self.trace.getStackTrace()
        return stack[index][1]

    def section(self, secname, filename=None):
        """
        Return the address of the section specified, if 
        filename is specified, use only that file to resolve
        the section name.

        Usage: section(<secname> [, <filename>])
        """
        if not filename:
            filename = self.trace.getExe()
        sym = self.trace.getSymByName(secname, filename)
        return long(sym)

    def sectionlen(self, secname, filename=None):
        """
        Return the length of the requested section. If not present
        the "libname" defaults to the primary executable.  If
        *present*, the libname must be a "normalized" library
        name (kernel32.dll == kernel32, /usr/lib/libfoo.1.30-3.so == libfoo).

        Usage: sectionlen(<secname> [,<libname>])
        """
        if not filename:
            filename = self.trace.getExe()

        sym = self.trace.getSymByName(secname, filename)
        return long(sym)

    def teb(self, threadnum=-1):
        """
        The expression teb(threadid) will return whatever the
        platform stores as the int for threadid.  In the case
        of windows, this is the TEB, others may be the thread
        stack base or whatever.  If threadid is left out, it
        uses the threadid of the current thread context.
        """
        if threadnum < 0:
            # Get the thread ID of the current Thread Context
            threadnum = self.trace.getMeta("ThreadId")

        teb = self.trace.getThreads().get(threadnum, None)
        if teb == None:
            raise Exception("ERROR - Unknown Thread Id %d" % threadnum)

        return teb

    def poi(self, address):
        """
        When expressions contain "poi(ebp)" this will return
        the address pointed to by ebp.
        """
        plen = struct.calcsize("P")
        mem = self.trace.readMemory(address, plen)
        return struct.unpack("P", mem)[0]

    def maplen(self, address):
        """
        The expression maplen(address) returns the length of the
        memory mapped area containing "address".
        """
        map = self.trace.getMap(address)
        if not map:
            raise Exception("ERROR - un-mapped address in maplen()")
        return map[1]

    def mapbase(self, address):
        """
        The expression mapbase(address) returns the base address of the
        memory mapped area containing "address"
        """
        map = self.trace.getMap(address)
        if not map:
            raise Exception("ERROR - un-mapped address in mapbase()")
        return map[0]

############################################################
# Platform mixins get thrown together here to form voltron
############################################################

def getTrace():
    """
    Return a tracer object appropriate for this platform.
    This is the function you will use to get a tracer object
    with the appropriate ancestry for your host.
    ex. mytrace = vtrace.getTrace()
    """

    if remote: #We have a remote server!
        return getRemoteTrace()

    import vtrace.platforms.base as v_base

    platform = sys.platform

    ilist = [object,] # Inheritors list
    ilist.append(Trace)
    ilist.append(v_base.BasePlatformMixin)
    ilist.append(v_base.UtilMixin)

    if sys.platform == "linux2":
        import vtrace.platforms.posix as v_posix
        import vtrace.platforms.linux as v_linux
        arch = os.uname()[4]
        ilist.append(v_posix.PosixMixin)
        ilist.append(v_posix.ElfMixin)
        if arch == "x86_64":
            #Order matters.
            import vtrace.archs.amd64 as v_amd64
            ilist.append(v_amd64.Amd64Mixin)
            ilist.append(v_posix.PtraceMixin)
            ilist.append(v_linux.LinuxMixin)
            ilist.append(v_linux.LinuxAmd64Registers)

        elif arch in ("i386","i486","i586","i686"):
            import vtrace.archs.intel as v_intel
            ilist.append(v_intel.IntelMixin)
            ilist.append(v_posix.PtraceMixin)
            ilist.append(v_linux.LinuxMixin)
            ilist.append(v_linux.LinuxIntelRegisters)

    elif sys.platform == "sunos5":
        print "SOLARIS SUPPORT ISNT DONE"
        import vtrace.platforms.posix as v_posix
        import vtrace.platforms.solaris as v_solaris
        arch = os.uname()[4]
        ilist.append(v_posix.PosixMixin)
        if arch == "i86pc":
            import vtrace.archs.intel as v_intel
            ilist.append(v_intel.IntelMixin)
            ilist.append(v_solaris.SolarisMixin)
            ilist.append(v_solaris.SolarisIntelMixin)

    elif sys.platform == "darwin":
        print "DARWIN SUPPORT ISNT DONE"
        if 9 not in os.getgroups():
            raise Exception("You must be in the procmod group!")
        import vtrace.platforms.darwin as v_darwin
        import vtrace.platforms.posix as v_posix
        arch = os.uname()[4]
        ilist.append(v_posix.PosixMixin)
        ilist.append(v_posix.PtraceMixin)
        ilist.append(v_darwin.DarwinMixin)
        ilist.append(v_darwin.MachoMixin)
        if arch == "i386":
            import vtrace.archs.intel as v_intel
            ilist.append(v_intel.IntelMixin)
            ilist.append(v_darwin.DarwinIntelRegisters)
        elif arch == "Power Macintosh":
            import vtrace.archs.intel as v_ppc
            ilist.append(v_ppc.PpcMixin)
            ilist.append(v_darwin.DarwinPpcRegisters)
        else:
            raise Exception("WTF?!?!  You got Darwin running on %s?!?>!?" % arch)

    elif sys.platform == "win32":
        import vtrace.platforms.win32 as v_win32
        import vtrace.archs.intel as v_intel
        ilist.append(v_win32.PEMixin)
        ilist.append(v_intel.IntelMixin)
        ilist.append(v_win32.Win32Mixin)

    else:
        raise Exception("ERROR - OS %s not supported yet" % platform)

    ilist.reverse()
    ttype = type("Trace", tuple(ilist), {})
    trace = ttype()

    # Lets let *any* mixin define a initMixin() routine
    # (mostly for initing modes)
    for cls in inspect.getmro(trace.__class__):
        if "initMixin" in dir(cls):
            cls.initMixin(trace)

    return trace

def interact(pid=0,server=None):

    """
    Just a cute and dirty way to get a tracer attached to a pid
    and get a python interpreter instance out of it.
    """

    global remote
    remote = server

    trace = getTrace()
    if pid:
        trace.attach(pid)

    mylocals = {}
    mylocals["trace"] = trace

    code.interact(local=mylocals)

