"""
Linux Platform Module
"""
# Copyright (C) 2007 Invisigoth - See LICENSE file for details
import os
import struct
import signal
import traceback

import vtrace
import vtrace.breakpoints as breakpoints
import vtrace.platforms.posix as v_posix
from vtrace.platforms.base import UtilMixin

import types

from ctypes import *

class GlibcNptlBreakpoint(vtrace.Breakpoint):
    """
    This breakpoint may be placed on any function that
    notifies the debugger of an NPTL event.
    """
    def __init__(self, libname, address):
        vtrace.Breakpoint.__init__(self, address)
        self.libname = libname
        
    def notify(self, event, trace):
        trace.nptlEvent(self.libname)

class GlibcNptlNotifier(vtrace.Notifier):
    """
    Hook the stuff in glibc's nptl implementation
    to get thread events and enumerate initial threads
    on attach.
    """
    def notify(self, event, trace):

        if event != vtrace.NOTIFY_LOAD_LIBRARY:
            return

        libfile = trace.getMeta("LatestLibrary", "")
        libnorm = trace.getMeta("LatestLibraryNorm", "")

        if ( libnorm.find("libpthread") == -1 and
            libfile != trace.getExe()):
            return

        try:
            nptl_version = trace.parseExpression("%s.nptl_version" % libnorm)
        except:
            return

        try:
            trace.nptlInitialize(libnorm)
        except Exception, msg:
            print "ERROR: NPTL Thread hooking failed on version:" , trace.readMemory(nptl_version, 20)
            print msg

class LinuxMixin:
    """
    The mixin to take care of linux specific platform traits.
    (mostly proc)
    """

    def initMixin(self):
        # Wrap reads from proc in our worker thread
        self.pthreads = [] # FIXME perhaps make this posix-wide not just linux eventually...
        self.threadWrap("platformReadMemory", self.platformReadMemory)
        self.threadWrap("doAttachThread", self.doAttachThread)
        self.setMeta("Platform", "Linux")
        self.nptlinit = False
        #FIXME - toying with the idea of "engine breakpoints" that the user
        # isn't aware of but the system can use to detect and fire events
        # of particular types...  This may be eventually part of the core...
        self.specialbreaks = {}

        self.registerNotifier(vtrace.NOTIFY_LOAD_LIBRARY, GlibcNptlNotifier())

    def checkBreakpoints(self):
        """
        Over-ride checkBreakpoints to hook the "special breaks" that are necissary
        for hooking thread creation and library loading...
        """
        if UtilMixin.checkBreakpoints(self):
            return True
        pc = self.getProgramCounter()
        bi = self.getBreakInstruction()
        bl = pc - len(bi)
        sbp = self.specialbreaks.get(bl, None)
        if sbp:
            self.setProgramCounter(bl)
            sbp.deactivate(self)
            sbp.notify(vtrace.NOTIFY_BREAK, self)
            # FIXME if "engine breaks" become part of the core,
            # un-ghettoify this
            self.stepping = True
            self.platformStepi()
            self.platformWait()
            self.stepping = False # FIXME this is a hack
            sbp.activate(self)
            self.setMeta("AutoContinue", True)
            return True
        return False

    def platformExec(self, cmdline):
        pid = v_posix.PtraceMixin.platformExec(self, cmdline)
        self.pthreads = [pid,]
        return pid

    def platformReadMemory(self, address, size):
        """
        A *much* faster way of reading memory that the 4 bytes
        per syscall allowed by ptrace
        """
        memfile = file("/proc/%d/mem" % self.pid, "r+b")
        memfile.seek(address)
        return memfile.read(size)

    def whynot_platformWriteMemory(self, address, data):
        """
        A *much* faster way of reading memory that the 4 bytes
        per syscall allowed by ptrace
        """
        memfile = file("/proc/%d/mem" % self.pid, "r+b")
        memfile.seek(address)
        return memfile.write(data)

    def platformAttach(self, pid):
        self.pthreads = [pid,]
        self.setMeta("ThreadId", pid)
        v_posix.PtraceMixin.platformAttach(self, pid)

    def platformPs(self):
        pslist = []
        for dname in os.listdir("/proc/"):
            try:
                if not dname.isdigit():
                    continue
                cmdline = file("/proc/%s/cmdline" % dname).read()
                cmdline = cmdline.replace("\x00"," ")
                if len(cmdline) > 0:
                    pslist.append((int(dname),cmdline))
            except:
                pass # Permissions...  Quick process... whatev.
        return pslist

    def attachThread(self, tid):
        self.doAttachThread(tid)
        self.setMeta("ThreadId", tid)
        self.fireNotifiers(vtrace.NOTIFY_CREATE_THREAD)

    def platformWait(self):
        # Blocking wait once...
        pid, status = os.waitpid(-1, 0x40000002)
        self.setMeta("ThreadId", pid)
        # Stop the rest of the threads... 
        # why is linux debugging so Ghetto?!?!
        if not self.stepping: # If we're stepping, only do the one
            for tid in self.pthreads:
                if tid == pid:
                    continue
                os.kill(tid, signal.SIGTRAP)
                os.waitpid(tid, 0x40000002)
        return status

    def platformContinue(self):
        cmd = v_posix.PT_CONTINUE
        if self.getMode("Syscall", False):
            cmd = v_posix.PT_SYSCALL
        pid = self.getPid()
        sig = self.getMeta("PendingSignal", 0)
        # Only deliver signals to the main thread
        if v_posix.ptrace(cmd, pid, 0, sig) != 0:
            raise Exception("ERROR ptrace failed for tid %d" % pid)

        for tid in self.pthreads:
            if tid == pid:
                continue
            if v_posix.ptrace(cmd, tid, 0, 0) != 0:
                pass

    def platformStepi(self):
        self.stepping = True
        tid = self.getMeta("ThreadId", 0)
        if v_posix.ptrace(v_posix.PT_STEP, tid, 0, 0) != 0:
            raise Exception("ERROR ptrace failed!")

    def platformDetach(self):
        for tid in self.pthreads:
            tid,v_posix.ptrace(v_posix.PT_DETACH, tid, 0, 0)

    def doAttachThread(self, tid):
        """
        Do the work for attaching a thread.  This must be *under*
        attachThread() so callers in notifiers may call it (because
        it's also gotta be thread wrapped).
        """
        if v_posix.ptrace(v_posix.PT_ATTACH, tid, 0, 0) != 0:
            raise Exception("ERROR ptrace attach failed for thread %d" % tid)
        os.waitpid(tid, 0x40000002)
        self.pthreads.append(tid)

    def platformGetRegs(self):
        x = (c_char * 512)()
        tid = self.getMeta("ThreadId", self.getPid())
        if v_posix.ptrace(v_posix.PT_GETREGS, tid, 0, addressof(x)) != 0:
            raise Exception("ERROR ptrace PT_GETREGS failed for TID %d" % tid)
        return x.raw

    def platformSetRegs(self, bytes):
        x = create_string_buffer(bytes)
        tid = self.getMeta("ThreadId")
        if v_posix.ptrace(v_posix.PT_SETREGS, tid, 0, addressof(x)) != 0:
            raise Exception("ERROR ptrace PT_SETREGS failed!")

    def threadsForPid(self, pid):
        ret = []
        for pidstr in os.listdir("/proc/%s/task" % pid):
            ret.append(int(pidstr))
        return ret

    def platformGetThreads(self):
        ret = {}
        for tid in self.pthreads:
            ret[tid] = tid #FIXME make this pthread struct or stackbase soon
        return ret

    def platformGetMaps(self):
        self.requireAttached()
        maps = []
        mapfile = file("/proc/%d/maps" % self.pid)
        for line in mapfile:

            perms = 0
            sline = line.split(" ")
            addrs = sline[0]
            permstr = sline[1]
            fname = sline[-1].strip()
            addrs = addrs.split("-")
            base = long(addrs[0],16)
            max = long(addrs[1],16)
            mlen = max-base

            if "r" in permstr:
                perms |= vtrace.MM_READ
            if "w" in permstr:
                perms |= vtrace.MM_WRITE
            if "x" in permstr:
                perms |= vtrace.MM_EXEC
            #if "p" in permstr:
                #pass

            maps.append((base,mlen,perms,fname))
        return maps

    def getExe(self):
        self.requireAttached()
        exe = os.readlink("/proc/%d/exe" % self.pid)
        if "(deleted)" in exe:
            if "#prelink#" in exe:
                exe = exe.split(".#prelink#")[0]
            elif ";" in exe:
                exe = exe.split(";")[0]
            else:
                exe = exe.split("(deleted)")[0].strip()
        return exe

    def platformGetFds(self):
        fds = []
        for name in os.listdir("/proc/%d/fd/" % self.pid):
            try:
                fdnum = int(name)
                fdtype = vtrace.FD_UNKNOWN
                link = os.readlink("/proc/%d/fd/%s" % (self.pid,name))
                if "socket:" in link:
                    fdtype = vtrace.FD_SOCKET
                elif "pipe:" in link:
                    fdtype = vtrace.FD_PIPE
                elif "/" in link:
                    fdtype = vtrace.FD_FILE

                fds.append((fdnum,fdtype,link))
            except:
                traceback.print_exc()

        return fds

############################################################
#
# These routines handle glibc's NPTL ABI
#
    def nptlOffset(self, libname, struc, field):
        addr = self.parseExpression("%s._thread_db_%s_%s" % (libname,struc,field))
        d1, d2, offset = self.readMemoryFormat(addr, "<3L")
        return offset

    def nptlEvent(self, libname):
        """
        Called when some thread somwhere has an event for us
        """
        TD_CREATE = 8 # From glibc headers...
        #FIXME make an NPTL context object to cache all this!
        last_event = self.parseExpression("%s.__nptl_last_event" % libname)
        nextevent = self.nptlOffset(libname, "pthread", "nextevent")
        eventbuf  = self.nptlOffset(libname, "pthread", "eventbuf")
        eventnum  = self.nptlOffset(libname, "td_eventbuf_t", "eventnum")
        eventdata = self.nptlOffset(libname, "td_eventbuf_t", "eventdata")
        tidoff    = self.nptlOffset(libname, "pthread","tid")

        pthread, = self.readMemoryFormat(last_event, "L")

        self.writeMemoryFormat(last_event, "L", 0)

        while pthread:
            pnext, = self.readMemoryFormat(pthread + nextevent, "L")
            self.writeMemoryFormat(pthread + nextevent, "L", 0)

            what, = self.readMemoryFormat(pthread + eventbuf + eventnum, "I")
            who,  = self.readMemoryFormat(pthread + eventbuf + eventdata, "L")

            if (what == TD_CREATE):
                tid, = self.readMemoryFormat(who + tidoff, "I")
                self.attachThread(tid)
                
            pthread = pnext


    def nptlGetThreads(self, libname):
        """
        Return the address of the pthread structs for all the threads
        """
        plist = self.parseExpression("%s.stack_used" % libname)
        puser = self.parseExpression("%s.__stack_user" % libname)
        listoff = self.nptlOffset(libname, "pthread","list")

        head,tail = self.readMemoryFormat(plist, "<2L")
        uhead,utail = self.readMemoryFormat(puser, "<2L")

        mainthread = uhead - listoff
        ret = [mainthread,]

        next = head

        while next != plist:
            ret.append(next - listoff)
            next,prev = self.readMemoryFormat(next, "<2L")

        return ret

    def nptlInitialize(self, libname):
        """
        Initialize glibc NPTL thread tracking
        """
        if self.nptlinit:
            return False

        pid = self.getPid()

        pevents = self.parseExpression("%s.__nptl_threads_events" % libname)

        create = self.parseExpression("%s.__nptl_create_event" % libname)
        tidoff = self.nptlOffset(libname, "pthread","tid")
        revoff = self.nptlOffset(libname, "pthread","report_events")
        evboff = self.nptlOffset(libname, "td_thr_events_t","event_bits")

        # Turn on all events!
        self.writeMemory(pevents + evboff, "\xff\xff\xff\xff")

        for pthread in self.nptlGetThreads(libname):
            # Tell this thread to report events
            self.writeMemory(pthread + revoff, "\xff\xff\xff\xff")
            tid, = self.readMemoryFormat(pthread + tidoff, "<L")
            if tid == pid:
                continue # We're already attached ;)
            self.attachThread(tid)

        bp = GlibcNptlBreakpoint(libname,create)
        bp.activate(self) # Activate by hand
        # Add this to the "special" breaks
        self.specialbreaks[bp.getAddress()] = bp

        self.nptlinit = True


class LinuxIntelRegisters:
    """
    The actual trace object for IntelLinux, which inherits
    what it can from IntelMixin/LinuxMixin and implements
    what is must.

    The size of the linux user area struct is 284 bytes...

    """

    def initMixin(self):
        self.usize = 284

    def maybeCalcUserSize(self):
        """
        LOL... this works actually... but we're not using it...
        """
        if self.usize != 0:
            return
        tid = self.getMeta("ThreadId", self.getPid())
        val = -1
        off = 500
        while val == -1:
            off -= 4
            val = v_posix.ptrace(v_posix.PT_READ_U, tid, off, 0)
            if off == 0:
                return
        self.usize = off + 4
        print "USIZE:",self.usize
        print "VAL:",val

    def platformGetRegs(self):
        """
        Start with what's given by PT_GETREGS and pre-pend
        the debug registers
        """
        tid = self.getMeta("ThreadId", self.getPid())
        buf = LinuxMixin.platformGetRegs(self)
        dbgs = []
        off = self.usize - 32
        for i in range(8):
            r = v_posix.ptrace(v_posix.PT_READ_U, tid, off+(4*i), 0)
            dbgs.append(r)

        return struct.pack("8L", *dbgs) + buf

    def platformSetRegs(self, buf):
        """
        Reverse of above...
        """
        v_posix.PtraceMixin.platformSetRegs(self, buf[32:])
        tid = self.getMeta("ThreadId", self.getPid())
        dbgs = struct.unpack("8L", buf[:32])
        off = self.usize - 32
        for i in range(8):
            v_posix.ptrace(v_posix.PT_WRITE_U, tid, off+(4*i), dbgs[i])

    def getRegisterFormat(self):
        return "15L8H2L2H2L2H"

    def getRegisterNames(self):
        return (
            "debug0","debug1","debug2","debug3","debug4","debug5",
            "debug6","debug7",
            "ebx","ecx","edx","esi","edi","ebp","eax","ds","__ds",
            "es","__es","fs","__fs","gs","__gs","orig_eax","eip",
            "cs","__cs","eflags","esp","ss","__ss")


class LinuxAmd64Registers:
    """
    Mixin for the register format on Linux AMD64
    """
    def getRegisterFormat(self):
        return "27L"

    def getRegisterNames(self):
        return (
            "r15","r14","r13","r12","rbp","rbx","r11","r10",
            "r9","r8","rax","rcx","rdx","rsi","rdi","orig_rax",
            "rip","cs","eflags",
            "rsp","ss",
            "fs_base"," gs_base",
            "ds","es","fs","gs")

